#include <gdnative_api_struct.gen.h>

#include <string.h>

#include "storybook.h"

const godot_gdnative_core_api_struct *api = NULL;
const godot_gdnative_ext_nativescript_api_struct *nativescript_api = NULL;

void *storybook_godot_constructor(godot_object *p_instance, void *p_method_data);
void storybook_godot_destructor(godot_object *p_instance, void *p_method_data, void *p_user_data);
godot_variant storybook_godot_parse(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args);
godot_variant storybook_godot_get_body(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args);
godot_variant storybook_godot_advance(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args);
godot_variant storybook_godot_get_footer(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args);

void GDN_EXPORT godot_gdnative_init(godot_gdnative_init_options *p_options) {
	api = p_options->api_struct;

	for (int i = 0; i < api->num_extensions; i++) {
		switch(api->extensions[i]->type) {
		case GDNATIVE_EXT_NATIVESCRIPT: {
			nativescript_api = (godot_gdnative_ext_nativescript_api_struct *)api->extensions[i];
		}; break;
		default: break;
		}
	}
}

void GDN_EXPORT godot_gdnative_terminate(godot_gdnative_terminate_options *p_options) {
	api = NULL;
	nativescript_api = NULL;
}

#define register_method(meth) \
	godot_instance_method meth = { NULL, NULL, NULL}; \
	meth.method = &storybook_godot_ ## meth; \
	nativescript_api->godot_nativescript_register_method(p_handle, "STORYBOOK", #meth, attributes, meth)

void GDN_EXPORT godot_nativescript_init(void *p_handle) {
	godot_instance_create_func create = { NULL, NULL, NULL };
	create.create_func = &storybook_godot_constructor;

	godot_instance_destroy_func destroy = { NULL, NULL, NULL };
	destroy.destroy_func = &storybook_godot_destructor;

	nativescript_api->godot_nativescript_register_class(p_handle, "STORYBOOK", "Reference",
	                                                    create, destroy);

	godot_method_attributes attributes = { GODOT_METHOD_RPC_MODE_DISABLED };

	register_method(parse);
	register_method(get_body);
	register_method(advance);
	register_method(get_footer);
}

typedef struct user_data_struct {
	Book *book;
} user_data_struct;

void *storybook_godot_constructor(godot_object *p_instance, void *p_method_data){
	user_data_struct *user_data = api->godot_alloc(sizeof(user_data_struct));
	return user_data;
}

void storybook_godot_destructor(godot_object *p_instance, void *p_method_data, void *p_user_data) {
	user_data_struct *user_data = (user_data_struct*) p_user_data;
	storybook_free_book(user_data->book);
	api->godot_free(user_data);
}

godot_variant storybook_godot_parse(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args) {
	godot_variant ret;
	godot_string str = api->godot_variant_as_string(p_args[0]);
	godot_char_string utf8 = api->godot_string_utf8(&str);

	user_data_struct *user_data = (user_data_struct*) p_user_data;
	user_data->book = storybook_make_book(api->godot_char_string_get_data(&utf8));
	api->godot_char_string_destroy(&utf8);

	api->godot_variant_new_nil(&ret);
	return ret;
}


godot_variant storybook_godot_get_body(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args) {
	godot_variant ret;
	godot_string str;
		
	user_data_struct *user_data = (user_data_struct*) p_user_data;
	char *body = storybook_get_body(user_data->book);
	api->godot_string_new(&str);
	api->godot_string_parse_utf8(&str, body);
	storybook_free_string(body);

	api->godot_variant_new_string(&ret, &str);
	api->godot_string_destroy(&str);

	return ret;
}

godot_variant storybook_godot_advance(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args) {
	godot_variant ret;

	user_data_struct *user_data = (user_data_struct*) p_user_data;

	if (p_num_args == 0) {
		storybook_advance_nooption(user_data->book);
	} else {
		uint64_t option = api->godot_variant_as_uint(p_args[0]);
		storybook_advance_option(user_data->book, option);
	}

	api->godot_variant_new_nil(&ret);
	return ret;
}
	
godot_variant storybook_godot_get_footer(godot_object *p_instance, void *p_method_data, void *p_user_data, int p_num_args, godot_variant **p_args) {
	godot_variant ret;

	user_data_struct *user_data = (user_data_struct*) p_user_data;
	footer ft = storybook_get_footer(user_data->book);

	api->godot_variant_new_int(&ret, ft);
	return ret;
}
